@@ -1,13 +1,13 @@
 from ._test_tree import TestTree
 from ._test_tree_browser import TestTreeBrowser
-from ._scorer import Scorer, DummyScorer, ClassifierScorer, GeneratorScorer
+from ._scorer import Scorer, DummyScorer, ClassifierScorer, GeneratorScorer, RawScorer
 from ._server import serve
-from ._embedding import embed
+from .embedders import _embed as embed
 from ._model import Model
 from ._topic_model import ChainTopicModel, StandardTopicModel
 from . import generators
 
-__version__ = '0.3.1'
+__version__ = '0.3.5'
 
 default_generators = {
     "abstract": TestTree(r"test_trees/abstract_capabilities.csv")
