@@ -1,10 +1,9 @@
 import pandas as pd
 import uuid
-import logging
-import random
 import os
 import io
 import time
+import re
 import numpy as np
 import pandas as pd
 from ._prompt_builder import PromptBuilder
@@ -13,7 +12,6 @@ from ._model import Model
 from ._topic_model import TopicLabelingModel, TopicMembershipModel
 import adatest
 from pathlib import Path
-import torch
 
 # class TestTreeIterator():
 #     def __init__(self, test_tree):
@@ -36,7 +34,7 @@ class TestTree():
     webserver. A TestTree object also conforms to most of the standard pandas DataFrame API.
     """
 
-    def __init__(self, tests=None, index=None, compute_embeddings=False, ensure_topic_markers=True, **kwargs):
+    def __init__(self, tests=None, labeling_model=TopicLabelingModel, membership_model=TopicMembershipModel, index=None, compute_embeddings=False, ensure_topic_markers=True, cache_file=None, **kwargs):
         """ Create a new test tree.
 
         Parameters
@@ -59,6 +57,9 @@ class TestTree():
         # the canonical ordered list of test tree columns
         column_names = ['topic', 'input', 'output', 'label', 'labeler', 'description']
 
+        self.labeling_model = labeling_model
+        self.membership_model = membership_model
+
         # create a new test tree in memory
         if tests is None:
             self._tests = pd.DataFrame([], columns=column_names, dtype=str)
@@ -74,6 +75,7 @@ class TestTree():
             self._tests_location = tests
             if os.path.isfile(tests) or isinstance(tests, io.TextIOBase):
                 self._tests = pd.read_csv(tests, index_col=0, dtype=str, keep_default_na=False)
+                self._tests.index = self._tests.index.map(str)
             else:
                 raise Exception(f"The provided tests file is not supported: {tests}")
 
@@ -93,6 +95,20 @@ class TestTree():
             self._tests['labeler'] = "dataset"
             self._tests['description'] = ''
 
+        elif isinstance(tests, list) and isinstance(tests[0], str):
+            self._tests = pd.DataFrame(columns=column_names)
+            self._tests['input'] = tests
+            self._tests['output'] = "[no output]"
+            self._tests['topic'] = ''
+            self._tests['label'] = ''
+            self._tests['labeler'] = ''
+            self._tests['description'] = ''
+            self._tests_location = None
+
+            if index is None:
+                index = [uuid.uuid4().hex for _ in range(len(tests))]
+            self._tests.index = index
+
         else:
             if index is None:
                 index = [uuid.uuid4().hex for _ in range(len(tests))]
@@ -124,6 +140,9 @@ class TestTree():
         # drop any duplicate index values
         self._tests = self._tests.groupby(level=0).first()
 
+        # fix spaces in topics names that are not URI encoded
+        self._tests['topic'] = self._tests['topic'].apply(lambda x: x.replace(" ", "%20"))
+
         # drop any duplicate rows
         self._tests.drop_duplicates(["topic", "input", "output", "labeler"], inplace=True)
 
@@ -145,6 +164,10 @@ class TestTree():
         # if self.auto_save:
         #     self._last_saved_tests = self._tests.copy()
 
+    @property
+    def name(self):
+        return re.split(r"\/", self._tests_location)[-1] if self._tests_location is not None else "Tests"
+
     def ensure_topic_markers(self):
         marked_topics = {t: True for t in set(self._tests.loc[self._tests["label"] == "topic_marker"]["topic"])}
         for topic in set(self._tests["topic"]):
@@ -190,6 +213,9 @@ class TestTree():
     def shape(self):
         return self._tests.shape
     @property
+    def str(self):
+        return self._tests.str
+    @property
     def iterrows(self):
         return self._tests.iterrows
     @property
@@ -208,7 +234,7 @@ class TestTree():
     def sort_values(self):
         return self._tests.sort_values
     
-    # NOTE: Can't delegate to df.append as it is depricated in favor of pd.concat, which we can't use due to type checks 
+    # NOTE: Can't delegate to df.append as it is deprecated in favor of pd.concat, which we can't use due to type checks 
     def append(self, test_tree, axis=0):
         if isinstance(test_tree, pd.DataFrame):
             self._tests = pd.concat([self._tests, test_tree], axis=axis)
@@ -247,6 +273,22 @@ class TestTree():
         ids = [id for id, test in self._tests.iterrows() if is_subtopic(topic, test.topic)]
         return self.loc[ids]
 
+    def topic_has_direct_tests(self, target_topic: str)-> bool:
+        """Check if a topic has direct tests."""
+        hdt_df = self._tests.apply(
+            lambda row: row['topic']==target_topic and row['label'] != 'topic_marker',
+            axis=1
+        )
+        return hdt_df.any()
+
+    def topic_has_subtopics(self, target_topic: str) -> bool:
+        """Check if a topic has subtopics."""
+        has_subtopics_df = self._tests.apply(
+            lambda row: row['topic']!=target_topic and is_subtopic(target_topic, row["topic"]),
+            axis=1
+        )
+        return has_subtopics_df.any()
+
     def adapt(self, scorer=None, generator=adatest.generators.OpenAI(), auto_save=False, user="anonymous", recompute_scores=False, drop_inactive_score_columns=False,
               max_suggestions=100, suggestion_thread_budget=0.5, prompt_builder=PromptBuilder(), active_generator="default", starting_path="",
               score_filter=-1e10, topic_model_scale=0): # TODO: remove active_generator and replace with the ability to set the generator?
@@ -331,21 +373,38 @@ class TestTree():
 
         already_seen = {}
         drop_ids = []
+
+        # catch duplicate tests in the same topic
         for id, test in self._tests.iterrows():
             k = test.topic + "|_ADA_JOIN_|" + test.input + "|_ADA_JOIN_|" + test.output
             if k in already_seen:
                 drop_ids.append(id)
             else:
                 already_seen[k] = True
-        self._tests = self._tests.drop(drop_ids, axis=0)
 
-    def _cache_embeddings(self):
-        """ Pre-compute the embeddings for all test cases in the test tree.
+        # see if any suggestions are duplicates of things already in the their topic
+        # (note we do this as a second loop so we know we have already marked all the
+        # members of the topic in already_seen)
+        for id, test in self._tests.iterrows():
+            if test.topic.endswith("/__suggestions__"):
+                k = test.topic[:-len("/__suggestions__")] + "|_ADA_JOIN_|" + test.input + "|_ADA_JOIN_|" + test.output
+                if k in already_seen:
+                    drop_ids.append(id)
+        self._tests.drop(drop_ids, axis=0, inplace=True)
+
+    def _cache_embeddings(self, ids=None):
+        """ Pre-compute the embeddings for the given test cases.
+
+        This is used so we can batch the computation don't compute them one at a time later.
         """
 
+        if ids is None:
+            ids = self._tests.index
+
         # see what new embeddings we need to compute
         all_strings = []
-        for id, test in self._tests.iterrows():
+        for id in ids:
+            test = self._tests.loc[id]
             if test.label == "topic_marker":
                 parts = test.topic.rsplit("/", 1)
                 str = parts[1] if len(parts) == 2 else ""
@@ -355,7 +414,7 @@ class TestTree():
                     all_strings.append(str)
 
         # suggestions topics don't have topic markers so we check for them separately
-        all_strings.append("__suggestions__")
+        # all_strings.append("__suggestions__")
         
         # we don't use the output of the embedding, just do this to get the embeddings cached
         adatest.embed(all_strings)
@@ -364,14 +423,15 @@ class TestTree():
         """ Impute missing labels in the test tree. """
         # TODO: this is just a random mock, it needs to implement real local topic models
 
-        self._cache_embeddings()
-
-        for id, test in self._tests.iterrows():
+        ids_to_impute = self._tests.index[self._tests["label"] == ""]
+        self._cache_embeddings(ids_to_impute)
+        for id in ids_to_impute:
+            test = self._tests.loc[id]
             if test.label == "":
                 if self.topic_membership_model(test.topic)(test.input) == "off_topic":
                     self._tests.loc[id, "label"] = "off_topic"
                 else:
-                    self._tests.loc[id, "label"] = self.topic_labeling_model(test.topic)(test.input, test.output)
+                    self._tests.loc[id, "label"] = "pass" if self.topic_labeling_model(test.topic)(test.input, test.output) < 0.5 else "fail"
                 self._tests.loc[id, "labeler"] = "imputed"
 
     # def predict_labels(self, topical_io_pairs):
@@ -415,20 +475,20 @@ class TestTree():
     def topic_labeling_model(self, topic):
         topic = topic.replace("/__suggestions__", "") # predict suggestions using their parent topic label model
         if topic not in self._topic_labeling_models:
-            self._topic_labeling_models[topic] = TopicLabelingModel(topic, self)
+            self._topic_labeling_models[topic] = self.labeling_model(topic, self)
         return self._topic_labeling_models[topic]
 
     def topic_membership_model(self, topic):
         topic = topic.replace("/__suggestions__", "") # predict suggestions using their parent topic membership model
         if topic not in self._topic_membership_models:
-            self._topic_membership_models[topic] = TopicMembershipModel(topic, self)
+            self._topic_membership_models[topic] = self.membership_model(topic, self)
         return self._topic_membership_models[topic]
 
     def retrain_topic_labeling_model(self, topic):
-        self._topic_labeling_models[topic] = TopicLabelingModel(topic, self)
+        self._topic_labeling_models[topic] = self.labeling_model(topic, self)
 
     def retrain_topic_membership_model(self, topic):
-        self._topic_membership_models[topic] = TopicMembershipModel(topic, self)
+        self._topic_membership_models[topic] = self.membership_model(topic, self)
 
     def drop_topic(self, topic):
         """ Remove a topic from the test tree. """
