@@ -1,16 +1,10 @@
-import time
-from IPython.display import display, HTML
-import openai
 import numpy as np
 import copy
-import math
 import pandas as pd
-import torch
 import json
 import re
-import collections
+from tqdm import tqdm
 
-from transformers import MODEL_FOR_NEXT_SENTENCE_PREDICTION_MAPPING
 from adatest.generators import TestTreeSource
 
 from .comm import JupyterComm
@@ -19,17 +13,12 @@ import pathlib
 import copy
 import re
 import logging
-import os
-import io
 import statistics
-import checklist
-import checklist.editor
 from threading import Timer
-from ._scorer import expand_template, clean_template, ClassifierScorer, GeneratorScorer, Scorer
-from ._prompt_builder import PromptBuilder
-import builtins
+from ._scorer import expand_template, clean_template, Scorer
 import adatest # Need to import like this to prevent circular dependencies
 import urllib.parse
+from .utils import is_subtopic
 
 # from https://gist.github.com/walkermatt/2871026
 def throttle(interval):
@@ -68,11 +57,8 @@ def file_log(*args):
     f.flush()
     f.close()
 
-def is_subtopic(topic, candidate):
-    # Returns true if candidate is a subtopic of topic
-    return True if re.search(r'^%s(/|$)' % topic.replace('+', r'\+'), candidate) else False
 
-def matches_filter(test, filter_text: str):
+def matches_filter(test, filter_text):
     if filter_text is None or filter_text == "":
         return True
     else:
@@ -178,7 +164,7 @@ class TestTreeBrowser():
         # ensure that each scorer's score column is in the test tree dataframe
         for c in self.score_columns:
             if c not in self.test_tree.columns:
-                self.test_tree[c] = ["" for _ in range(self.test_tree.shape[0])]
+                self.test_tree[c] = [np.nan if label == "topic_marker" else "__TOEVAL__" for label in self.test_tree["label"]]
 
         # a unique identifier for this test set instance, used for UI connections
         self._id = uuid.uuid4().hex
@@ -215,6 +201,37 @@ class TestTreeBrowser():
         # self.suggestions = pd.DataFrame([], columns=self.test_tree.columns)
         self._suggestions_error = "" # tracks if we failed to generate suggestions
 
+    def auto_optimize(self, rounds=10, topic=""):
+        """ Run the testing loop for a topic without user involvement.
+        
+        Note that this assumes the labeling model is always correct.
+        """
+
+        for _ in tqdm(list(range(rounds))):
+    
+            # create new suggestions in the topic
+            self.generate_suggestions(topic)
+            
+            # get the ids of the on-topic suggestions
+            keep_ids = []
+            drop_ids = []
+            for k, test in self.test_tree.iterrows():
+                main_score = test[self.score_columns[0]]
+                if test.topic == topic+"/__suggestions__":
+                    if test.label != "off_topic" and not isinstance(main_score, str) and not np.isnan(main_score):
+                        keep_ids.append(k)
+                    else:
+                        drop_ids.append(k)
+            
+            # print(tests.loc[top10_ids[0], "model score"])
+            # print(tests.loc[top10_ids[0], "input"])
+            # print()
+            
+            # label and move these top suggestions to the root topic
+            self.test_tree.loc[keep_ids, "labeler"] = "auto_optimize"
+            self.test_tree.loc[keep_ids, "topic"] = topic
+            self.test_tree.drop(drop_ids, inplace=True)
+
     def _repr_html_(self, prefix="", environment="jupyter", websocket_server=None):
         """ Returns the HTML interface for this browser.
 
@@ -253,6 +270,7 @@ class TestTreeBrowser():
     def display(self):
         """ Manually display the HTML interface.
         """
+        from IPython.display import display, HTML
         display(HTML(self._repr_html_()))
 
     def interface_event(self, msg):
@@ -267,212 +285,215 @@ class TestTreeBrowser():
 
         log.debug(f"interface_event({msg})")
 
-        # loop over each event message
-        for k in msg:
-            if k == "browser":
-                action = msg[k].get("action", None)
-                
-                # rewdraw the entire interface
-                if action == "redraw":
-                    self._refresh_interface()
-                
-                # generate a new set of suggested tests/topics
-                elif action == "generate_suggestions":
-                    self._clear_suggestions()
-                    self.test_tree.retrain_topic_labeling_model(self.current_topic)
-                    self.test_tree.retrain_topic_membership_model(self.current_topic)
-                    self._generate_suggestions(filter=msg[k].get("filter", ""))
-                    # if self._active_generator_obj is None:
-                    #     self._suggestions_error = "No AdaTest generator has been set!"
-                    # else:
-                    #     self._generate_suggestions(filter=msg[k].get("filter", ""))
-                        # # try:
-                        # self.suggestions = self._generate_suggestions(filter=msg[k].get("filter", ""))
-                        # # filter suggestions to relevant types
-                        # if self.mode == "topics":
-                        #     self.suggestions = self.suggestions[self.suggestions['type'] == "topic_marker"]
-                        # elif self.mode == "tests":
-                        #     self.suggestions = self.suggestions[self.suggestions['type'] != "topic_marker"]
+        if "event_id" not in msg:
+            log.error("interface_event: missing event_id. msg dump: %s", msg)
+            return
+        event_id = msg["event_id"]
 
-                        # # Ensure valid suggestions exist.
-                        # if self.suggestions.shape[0] > 0:  
-                        #     self.suggestions.sort_values(self.score_columns[0], inplace=True, ascending=False, key=np.vectorize(score_max))
-                        #     self._suggestions_error = ""
-                        # else:
-                        #     self._suggestions_error = True # Not sure if we should do this?
-                        # except Exception as e:
-                        #     log.debug(e)
-                        #     self.suggestions = pd.DataFrame([], columns=self.test_tree.columns)
-                        #     self._suggestions_error = True
-                    self._refresh_interface()
-                
-                # change the current topic
-                elif action == "change_topic":
-                    self.current_topic = msg[k]["topic"]
-                    # self.suggestions = pd.DataFrame([], columns=self.test_tree.columns)
+        # redraw the entire interface
+        if event_id == "redraw":
+            self._refresh_interface()
 
-                    # see if we have only topics are direct children, if so, we suggest topics, otherwise we suggest tests
-                    has_direct_tests = False
-                    has_known_subtopics = False
-                    for k, test in self.test_tree.iterrows():
-                        if test["topic"] == self.current_topic:
-                            if test["label"] != "topic_marker":
-                                has_direct_tests = True
-                        elif is_subtopic(self.current_topic, test["topic"]):
-                            has_known_subtopics = True
-                    if not has_direct_tests and has_known_subtopics:
-                        self.mode = "topics"
-                    else:
-                        self.mode = "tests"
+        # generate a new set of suggested tests/topics
+        elif event_id == "generate_suggestions":
+            self._clear_suggestions()
+            self.test_tree.retrain_topic_labeling_model(self.current_topic)
+            self.test_tree.retrain_topic_membership_model(self.current_topic)
+            self._generate_suggestions(filter=msg.get("filter", ""))
+            # if self._active_generator_obj is None:
+            #     self._suggestions_error = "No AdaTest generator has been set!"
+            # else:
+            #     self._generate_suggestions(filter=msg[k].get("filter", ""))
+            # # try:
+            # self.suggestions = self._generate_suggestions(filter=msg[k].get("filter", ""))
+            # # filter suggestions to relevant types
+            # if self.mode == "topics":
+            #     self.suggestions = self.suggestions[self.suggestions['type'] == "topic_marker"]
+            # elif self.mode == "tests":
+            #     self.suggestions = self.suggestions[self.suggestions['type'] != "topic_marker"]
 
-                    self._refresh_interface()
-                
-                # clear the current set of suggestions
-                elif action == "clear_suggestions":
-                    self._clear_suggestions()
-                    # self.suggestions = pd.DataFrame([], columns=self.test_tree.columns)
-                    self._refresh_interface()
-
-                # add a new empty subtopic to the current topic
-                elif action == "add_new_topic":
-                    self.test_tree.loc[uuid.uuid4().hex] = {
-                        "topic": self.current_topic + "/New topic",
-                        "label": "topic_marker",
-                        "input": "",
-                        "output": "",
-                        "labeler": self.user,
-                        "description": ""
-                    }
-                    self._compute_embeddings_and_scores(self.test_tree)
-                    self._auto_save()
-                    self._refresh_interface()
-                
-                # add a new empty test to the current topic
-                elif action == "add_new_test":
-
-                    # add the new test row
-                    row = {
-                        "topic": self.current_topic,
-                        "input": "New test", # The special value "New test" causes the interface to auto-select the text
-                        "output": "",
-                        "label": "",
-                        "labeler": "imputed",
-                        "description": ""
-                    }
-                    for c in self.score_columns:
-                        row[c] = "__TOEVAL__"
-                        row[c[:-6] + " raw outputs"] = "{}"
-                    self.test_tree.loc[uuid.uuid4().hex] = row
-
-                    self._compute_embeddings_and_scores(self.test_tree)
-                    self._auto_save()
-                    self._refresh_interface()
-
-                # change which scorer/model is used for sorting tests
-                elif action == "set_first_model":
-                    name = msg[k]["model"]
-
-                    # move to front of score columns
-                    pos = len(self.test_tree.columns) - len(self.score_columns)
-                    tmp = self.test_tree[name]
-                    self.test_tree.drop(labels=[name], axis=1, inplace=True)
-                    self.test_tree.insert(pos, name, tmp)
-
-                    # update score columns list
-                    self.score_columns.remove(name)
-                    self.score_columns.insert(0, name)
-
-                    self._auto_save()
-                    self._refresh_interface()
-
-                # change which generator is active
-                elif action is None and "active_generator" in msg[k]:
-                    self.active_generator = msg[k]["active_generator"]
-                    self._active_generator_obj = self.generators[self.active_generator]
-
-                # change between topics and tests
-                elif action is None and "mode" in msg[k]:
-                    self.mode = msg[k]["mode"]
-
-                elif action == 'change_description':
-                    id = msg[k]['topic_marker_id']
-                    if id not in self.test_tree.index:
-                        self.test_tree.loc[id, 'topic'] = "" # only the root topic would be missing from the tree
-                        self.test_tree.loc[id, 'input'] = ""
-                        self.test_tree.loc[id, 'output'] = ""
-                        self.test_tree.loc[id, 'label'] = "topic_marker"
-                    self.test_tree.loc[id, 'description'] = msg[k]['description']
-                    self._auto_save()
-
-                elif action == 'change_filter':
-                    print("change_filter")
-                    self.filter_text = msg[k]['filter_text']
-                    self._refresh_interface()
-
-
-            # if we are just updating a single row in tests then we only recompute the scores
-            elif "topic" not in msg[k]:
-                sendback_data = {}
-                
-                # convert template expansions into a standard value update
-                if msg[k].get("action", "") == "template_expand":
-                    template_value = self.templatize(self.test_tree.loc[k, msg[k]["value"]])
-                    msg[k] = {msg[k]["value"]: template_value}
-                    sendback_data[msg[k]["value"]] = template_value
-
-                # update the row and recompute scores
-                for k2 in msg[k]:
-                    self.test_tree.loc[k, k2] = msg[k][k2]
-                if "input" in msg[k] or "output" in msg[k]:
-                    self.test_tree.loc[k, self.score_columns] = "__TOEVAL__"
-                    self._compute_embeddings_and_scores(self.test_tree, overwrite_outputs="output" not in msg[k])
-                elif "label" in msg[k]:
-                    #self.test_tree.retrain_topic_model(self.current_topic)
-                    pass # SML: we could recompute the scores here but then that would change the output of stochastic output models
-                    # sign = -1 if msg[k]["label"] == "pass" else 1
-                    # self.test_tree.loc[k, self.score_columns] = ""#abs(float(self.test_tree.loc[k, self.score_columns])) #* sign
-                    # self._compute_embeddings_and_scores(self.test_tree, overwrite_outputs=False)
-
-                # send just the data that changed back to the frontend
-                sendback_data["scores"] = {c: [[k, v] for v in ui_score_parts(self.test_tree.loc[k, c], self.test_tree.loc[k, "label"])] for c in self.score_columns}
-                outputs = {c: [[k, json.loads(self.test_tree.loc[k].get(c[:-6] + " raw outputs", "{}"))]] for c in self.score_columns}
-                sendback_data["raw_outputs"] = outputs
-                if "output" not in msg[k]: # if the output was given to us the client is managing its current state so we shouldn't send it back
-                    sendback_data["output"] = self.test_tree.loc[k, "output"]
-                sendback_data["label"] = self.test_tree.loc[k, "label"]
-                sendback_data["labeler"] = self.test_tree.loc[k, "labeler"]
-                sendback_data.update(self.test_display_parts(self.test_tree.loc[k]))
-                self.comm.send({k: sendback_data})
-                
-                self._auto_save()
-
-            # if we are just changing the topic
-            elif "topic" in msg[k] and len(msg[k]) == 1:
-                
-                # move a test that is in the test tree
-                if k in self.test_tree.index:
-                    if msg[k]["topic"] == "_DELETE_": # this means delete the test
-                        self.test_tree.drop(k, inplace=True)
-                    else:
-                        self.test_tree.loc[k, "topic"] = msg[k]["topic"]
-                        self.test_tree.loc[k, "author"] = self.user
-                
-                # move a whole topic around
-                else:
-                    for id, test in self.test_tree.iterrows():
-                        if is_subtopic(k, test.topic):
-                            if msg[k]["topic"] == "_DELETE_":
-                                self.test_tree.drop(id, inplace=True)
-                            else:
-                                self.test_tree.loc[id, "topic"] = msg[k]["topic"] + test.topic[len(k):]
-
-                # Recompute any missing embeddings to handle any changes
-                self._compute_embeddings_and_scores(self.test_tree)
-                self._refresh_interface()
-                self._auto_save()
+            # # Ensure valid suggestions exist.
+            # if self.suggestions.shape[0] > 0:  
+            #     self.suggestions.sort_values(self.score_columns[0], inplace=True, ascending=False, key=np.vectorize(score_max))
+            #     self._suggestions_error = ""
+            # else:
+            #     self._suggestions_error = True # Not sure if we should do this?
+            # except Exception as e:
+            #     log.debug(e)
+            #     self.suggestions = pd.DataFrame([], columns=self.test_tree.columns)
+            #     self._suggestions_error = True
+            self._refresh_interface()
+            
+        # change the current topic
+        elif event_id == "change_topic":
+            self.current_topic = msg["topic"]
+            # self.suggestions = pd.DataFrame([], columns=self.test_tree.columns)
 
+            # see if we have only topics are direct children, if so, we suggest topics, otherwise we suggest tests
+            has_direct_tests = self.test_tree.topic_has_direct_tests(self.current_topic)
+            has_known_subtopics = self.test_tree.topic_has_subtopics(self.current_topic)
+            if not has_direct_tests and has_known_subtopics:
+                self.mode = "topics"
             else:
-                log.debug(f"Unable to parse the interface message: {msg[k]}")
+                self.mode = "tests"
+
+            self._refresh_interface()
+            
+        # clear the current set of suggestions
+        elif event_id == "clear_suggestions":
+            self._clear_suggestions()
+            # self.suggestions = pd.DataFrame([], columns=self.test_tree.columns)
+            self._refresh_interface()
+
+        # add a new empty subtopic to the current topic
+        elif event_id == "add_new_topic":
+            self.test_tree.loc[uuid.uuid4().hex] = {
+                "topic": self.current_topic + "/New topic",
+                "label": "topic_marker",
+                "input": "",
+                "output": "",
+                "labeler": self.user,
+                "description": ""
+            }
+            self._compute_embeddings_and_scores(self.test_tree)
+            self._auto_save()
+            self._refresh_interface()
+            
+        # add a new empty test to the current topic
+        elif event_id == "add_new_test":
+
+            # add the new test row
+            row = {
+                "topic": self.current_topic,
+                "input": "New test", # The special value "New test" causes the interface to auto-select the text
+                "output": "",
+                "label": "",
+                "labeler": "imputed",
+                "description": ""
+            }
+            for c in self.score_columns:
+                row[c] = np.nan
+                row[c[:-6] + " raw outputs"] = "{}"
+            self.test_tree.loc[uuid.uuid4().hex] = row
+
+            self._auto_save()
+            self._refresh_interface()
+
+        # change which scorer/model is used for sorting tests
+        elif event_id == "set_first_model":
+            name = msg["model"]
+
+            # move to front of score columns
+            pos = len(self.test_tree.columns) - len(self.score_columns)
+            tmp = self.test_tree[name]
+            self.test_tree.drop(labels=[name], axis=1, inplace=True)
+            self.test_tree.insert(pos, name, tmp)
+
+            # update score columns list
+            self.score_columns.remove(name)
+            self.score_columns.insert(0, name)
+
+            self._auto_save()
+            self._refresh_interface()
+
+        elif event_id == "change_generator":
+            self.active_generator = msg["generator"]
+            self._active_generator_obj = self.generators[self.active_generator]
+
+        elif event_id == "change_mode":
+            self.mode = msg["mode"]
+
+        elif event_id == 'change_description':
+            id = msg['topic_marker_id']
+            if id not in self.test_tree.index:
+                self.test_tree.loc[id, 'topic'] = "" # only the root topic would be missing from the tree
+                self.test_tree.loc[id, 'input'] = ""
+                self.test_tree.loc[id, 'output'] = ""
+                self.test_tree.loc[id, 'label'] = "topic_marker"
+            self.test_tree.loc[msg['topic_marker_id'], 'description'] = msg['description']
+            self._auto_save()
+            self._refresh_interface()
+
+        elif event_id == 'change_filter':
+            print("change_filter")
+            self.filter_text = msg['filter_text']
+            self._refresh_interface()
+
+        # Move a test/topic to a new topic
+        # Also used to rename
+        elif event_id == "move_test":
+            log.debug("move_test")
+            test_ids = msg["test_ids"]
+            # test_id can either be a unique test ID or a topic name
+            for test_id in test_ids:
+                if test_id in self.test_tree.index:
+                    self.test_tree.loc[test_id, "topic"] = msg["topic"]
+                    self.test_tree.loc[test_id, "author"] = self.user
+                if '/' in test_id:
+                    for id, test in self.test_tree.iterrows():
+                        if is_subtopic(test_id, test.topic):
+                            self.test_tree.loc[id, "topic"] = msg["topic"] + test.topic[len(test_id):]
+            # Recompute any missing embeddings to handle any changes
+            self._compute_embeddings_and_scores(self.test_tree)
+            self._auto_save()
+            self._refresh_interface()
+
+        elif event_id == "delete_test":
+            log.debug("delete_test")
+            test_ids = msg["test_ids"]
+            # test_id can either be a unique test ID or a topic name
+            for test_id in test_ids:
+                if test_id in self.test_tree.index:
+                    self.test_tree.drop(test_id, inplace=True)
+                if '/' in test_id:
+                    # Delete tests from subtopics
+                    for id, test in self.test_tree.iterrows():
+                        if is_subtopic(test_id, test.topic):
+                            self.test_tree.drop(id, inplace=True)
+            self._compute_embeddings_and_scores(self.test_tree)
+            self._auto_save()
+            self._refresh_interface()
+        
+        # if we are just updating a single row in tests then we only recompute the scores
+        elif event_id == "change_label" or event_id == "change_input" or event_id == "change_output":
+            sendback_data = {}
+            test_id = msg["test_ids"][0]
+            
+            # convert template expansions into a standard value update
+            if msg.get("action", "") == "template_expand":
+                template_value = self.templatize(self.test_tree.loc[test_id, msg["value"]])
+                msg = {msg["value"]: template_value}
+                sendback_data[msg["value"]] = template_value
+
+            # update the row and recompute scores
+            metadata_fields = ["event_id", "test_ids"]
+            for k2 in msg:
+                if k2 not in metadata_fields:
+                    self.test_tree.loc[test_id, k2] = msg[k2]
+            if event_id == "change_input":
+                self.test_tree.loc[test_id, self.score_columns] = "__TOEVAL__"
+                self._compute_embeddings_and_scores(self.test_tree, overwrite_outputs="output" not in msg)
+            elif event_id == "change_label":
+                # sign = -1 if msg["label"] == "pass" else 1
+                # self.test_tree.loc[test_id, self.score_columns] = abs(float(self.test_tree.loc[test_id, self.score_columns])) * sign
+                pass # SML: we could recompute the scores here but then that would change the output of stochastic output models
+                # self._compute_embeddings_and_scores(self.test_tree, overwrite_outputs=False)
+
+            # send just the data that changed back to the frontend
+            sendback_data["scores"] = {c: [[test_id, v] for v in ui_score_parts(self.test_tree.loc[test_id, c], self.test_tree.loc[test_id, "label"])] for c in self.score_columns}
+            outputs = {c: [[test_id, json.loads(self.test_tree.loc[test_id].get(c[:-6] + " raw outputs", "{}"))]] for c in self.score_columns}
+            sendback_data["raw_outputs"] = outputs
+            if "output" not in msg: # if the output was given to us the client is managing its current state so we shouldn't send it back
+                sendback_data["output"] = self.test_tree.loc[test_id, "output"]
+            sendback_data["label"] = self.test_tree.loc[test_id, "label"]
+            sendback_data["labeler"] = self.test_tree.loc[test_id, "labeler"]
+            sendback_data.update(self.test_display_parts(self.test_tree.loc[test_id]))
+            self.comm.send({test_id: sendback_data})
+            
+            self._auto_save()
+        
+        else:
+            log.error(f"Unable to parse the interface message: {msg}")
 
     def _refresh_interface(self):
         """ Send our entire current state to the frontend interface.
@@ -485,44 +506,46 @@ class TestTreeBrowser():
             children = []
             
             # add tests and topics to the data lookup structure
-            for k, test in tests.iterrows():
-                if is_subtopic(topic, test.topic):
+            subtopic_ids = tests.index[tests["topic"].str.match(r"^%s(/|$)" % re.escape(topic))]
+            for k in subtopic_ids:
+                test = tests.loc[k]
                     
-                    # add a topic
-                    if test.label == "topic_marker":
-                        if is_subtopic(topic, test.topic) and test.topic != topic:
-                            name = test.topic[len(topic)+1:]
-                            if "/" not in name: # only add direct children
-                                data[test.topic] = {
-                                    "label": test.label,
-                                    "labeler": test.labeler,
-                                    "description": "",
-                                    "scores": {c: [] for c in self.score_columns},
-                                    "topic_marker_id": k,
-                                    "topic_name": name,
-                                    "editing": test.topic.endswith("/New topic")
-                                }
-                                children.append(test.topic)
-                    
-                    # add a test
-                    elif matches_filter(test, self.filter_text):
-                        data[k] = {
-                            "input": test.input,
-                            "output": test.output,
-                            "label": test.label,
-                            "labeler": test.labeler,
-                            "description": test.description,
-                            "scores": {c: [[k, v] for v in ui_score_parts(test[c], test.label)] for c in self.score_columns},
-                            "editing": test.input == "New test"
-                        }
+                # add a topic
+                if test.label == "topic_marker":
+                    if test.topic != topic:
+                        name = test.topic[len(topic)+1:]
+                        if "/" not in name: # only add direct children
+                            data[test.topic] = {
+                                "label": test.label,
+                                "labeler": test.labeler,
+                                "description": "",
+                                "scores": {c: [] for c in self.score_columns},
+                                "topic_marker_id": k,
+                                "topic_name": name,
+                                "editing": test.topic.endswith("/New topic")
+                            }
+                            children.append(test.topic)
+                
+                # add a test
+                elif matches_filter(test, self.filter_text):
+                    data[k] = {
+                        "input": test.input,
+                        "output": test.output,
+                        "label": test.label,
+                        "labeler": test.labeler,
+                        "description": test.description,
+                        "scores": {c: [[k, v] for v in ui_score_parts(test[c], test.label)] for c in self.score_columns},
+                        "editing": test.input == "New test"
+                    }
 
-                        data[k]["raw_outputs"] = {c: [[k, safe_json_load(test.get(c[:-6] + " raw outputs", "{}"))]] for c in self.score_columns}
-                        data[k].update(self.test_display_parts(test))
-                        if test.topic == topic:
-                            children.append(k)
+                    data[k]["raw_outputs"] = {c: [[k, safe_json_load(test.get(c[:-6] + " raw outputs", "{}"))]] for c in self.score_columns}
+                    data[k].update(self.test_display_parts(test))
+                    if test.topic == topic:
+                        children.append(k)
             
             # fill in the scores for the child topics
-            for k, test in tests.iterrows():
+            for k in subtopic_ids:
+                test = tests.loc[k]
                 if "/__suggestions__" not in test.topic and is_subtopic(topic, test.topic) and test.topic != topic:
                     child_topic = test.topic[len(topic):].split("/", 2)[1]
                     scores = data[topic+"/"+child_topic]["scores"]
@@ -599,6 +622,7 @@ class TestTreeBrowser():
             "active_generator": self.active_generator,
             "mode": self.mode,
             "mode_options": self.mode_options,
+            "test_tree_name": self.test_tree.name
             # "test_types": test_types,
             # "test_type_parts": test_type_parts,
         }
@@ -704,7 +728,7 @@ class TestTreeBrowser():
                     id = uuid.uuid4().hex
                     self.test_tree.loc[id, "topic"] = self.current_topic + "/__suggestions__" + ("/"+input if self.mode == "topics" else "")
                     self.test_tree.loc[id, "input"] = "" if self.mode == "topics" else input
-                    self.test_tree.loc[id, "output"] = "__TOOVERWRITE__"
+                    self.test_tree.loc[id, "output"] = "[no output]"
                     self.test_tree.loc[id, "label"] = "topic_marker" if self.mode == "topics" else ""
                     self.test_tree.loc[id, "labeler"] = "imputed"
                     self.test_tree.loc[id, "description"] = ""
@@ -726,6 +750,10 @@ class TestTreeBrowser():
                         test_map_tmp[str_val] = True
 
             # suggestions = pd.DataFrame(suggestions, index=[uuid.uuid4().hex for _ in range(len(suggestions))], columns=self.test_tree.columns)
+            # make sure any duplicates we may have introduced are removed
+            self.test_tree.deduplicate()
+            
+            # compute the scores for the new tests
             self._compute_embeddings_and_scores(self.test_tree)
 
         # Filter invalid suggestions
@@ -756,6 +784,9 @@ class TestTreeBrowser():
         return topic_marker_index
 
     def _add_topic_model_score(self, df, topic_model_scale):
+        """ This is an old experimental funciton that is not meant to be used anymore.
+        """
+        import openai
         documents = []
         for k,s in df.iterrows():
             max_output = -10e8
@@ -787,12 +818,17 @@ class TestTreeBrowser():
     def _compute_embeddings_and_scores(self, tests, recompute=False, overwrite_outputs=False, save_outputs=False): # TODO: Rename/refactor/merge with _compute_scores?
         log.debug(f"compute_embeddings_and_scores(tests=<DataFrame shape={tests.shape}>, recompute={recompute})")
 
+        # nothing to do if we don't have a scorer
+        if self.scorer is None:
+            return
+        
         for k in self.scorer:
             # determine which rows we need to evaluate
-            eval_ids = []
-            for i, (id, test) in enumerate(tests.iterrows()):
-                if (recompute or test[k+" score"] == "__TOEVAL__") and test.label != "topic_marker" and test.label != "off_topic":
-                    eval_ids.append(id)
+            # eval_ids = []
+            # for i, (id, test) in enumerate(tests.iterrows()):
+            #     if (recompute or test[k+" score"] == "__TOEVAL__" or test["output"] == "[no output]") and test.label != "topic_marker" and test.label != "off_topic":
+            #         eval_ids.append(id)
+            eval_ids = tests.index[((tests[k+" score"] == "__TOEVAL__") | (tests["output"] == "[no output]")) & (tests["label"] != "topic_marker") & (tests["label"] != "off_topic")]
 
             if len(eval_ids) > 0:
 
@@ -804,7 +840,7 @@ class TestTreeBrowser():
                 for i,id in enumerate(eval_ids):
                     # tests.loc[id, k+" score"] = scores[i]
 
-                    if not overwrite_outputs and current_outputs.loc[id] != "__TOOVERWRITE__" and current_outputs.loc[id] != new_outputs[i]:
+                    if not overwrite_outputs and current_outputs.loc[id] != "[no output]" and current_outputs.loc[id] != new_outputs[i]:
 
                         # mark the current row as nan score (meaning the output does not match)
                         tests.loc[id, k+" score"] = np.nan
@@ -822,7 +858,8 @@ class TestTreeBrowser():
                         tests.loc[id, "output"] = new_outputs[i]
                         tests.loc[id, k+" score"] = scores[i]
 
-        tests.deduplicate() # make sure any duplicates we may have introduced are removed
+        # make sure any duplicates we may have introduced are removed
+        # tests.deduplicate()
 
         # reimpute missing labels
         tests.impute_labels() # TODO: ensure this method caches the local models and only reimputes when needed for each topic
@@ -1017,6 +1054,9 @@ class TestTreeBrowser():
         }
 
     def templatize(self, s):
+        """ This is an experimental function that is not meant to be used generally.
+        """
+        import openai
         prompt = """INPUT: "Where are regular people on Twitter"
     OUTPUT: "Where are {regular|normal|sane|typical} people on {Twitter|Facebook|Reddit|Instagram}"
     ###
